==> birda-1.1.tar.gz: birda-1.1/src/sir.c <==
/*
 * Copyright (c) 2001 Tommy Bohlin <tommy@gatespace.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/* sir.c
 */

#include <irda.h>

#include <stdlib.h>

/**********************************************************************
 * Constants
 **********************************************************************/

static const char id_sir_framedevice[]="sir frame device";
static const char id_sir_inbuffer[]="sir inbuffer";

/**********************************************************************
 * Data structures
 **********************************************************************/

#define SIR_EXTRA_BOF            0xff
#define SIR_BOF                  0xc0
#define SIR_CE                   0x7d
#define SIR_EOF                  0xc1

#define SIR_ESC_BIT              0x20

typedef struct SIRFrameDevice {
  FrameDevice frame;
  SerialDevice* serial;

  u_char* inBuf;
  int maxSize;
  int frameState;
#define FRAME_OUTSIDE    0
#define FRAME_INSIDE     1
#define FRAME_ESCAPE     2
  int inChars;
  int inFCS;

  int ebofs;
} SIRFrameDevice;

/**********************************************************************
 * CRC
 **********************************************************************/

static const int fcstab[] = {
  0x0000, 0x1189, 0x2312, 0x329b, 0x4624, 0x57ad, 0x6536, 0x74bf,
  0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5, 0xe97e, 0xf8f7,
  0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7, 0x643e,
  0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
  0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd,
  0xad4a, 0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5,
  0x3183, 0x200a, 0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c,
  0xbdcb, 0xac42, 0x9ed9, 0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974,
  0x4204, 0x538d, 0x6116, 0x709f, 0x0420, 0x15a9, 0x2732, 0x36bb,
  0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868, 0x99e1, 0xab7a, 0xbaf3,
  0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528, 0x37b3, 0x263a,
  0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb, 0xaa72,
  0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
  0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1,
  0x7387, 0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738,
  0xffcf, 0xee46, 0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70,
  0x8408, 0x9581, 0xa71a, 0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7,
  0x0840, 0x19c9, 0x2b52, 0x3adb, 0x4e64, 0x5fed, 0x6d76, 0x7cff,
  0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad, 0xc324, 0xf1bf, 0xe036,
  0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c, 0x7df7, 0x6c7e,
  0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c, 0xd1b5,
  0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
  0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134,
  0x39c3, 0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c,
  0xc60c, 0xd785, 0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3,
  0x4a44, 0x5bcd, 0x6956, 0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb,
  0xd68d, 0xc704, 0xf59f, 0xe416, 0x90a9, 0x8120, 0xb3bb, 0xa232,
  0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1, 0x0d68, 0x3ff3, 0x2e7a,
  0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3, 0x8238, 0x93b1,
  0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70, 0x1ff9,
  0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
  0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78
};

static const int INITFCS=0xffff;
static const int GOODFCS=0xf0b8;

static int updateFCS(int fcs, int c) {
  return (fcs>>8)^fcstab[0xff&(fcs^c)];
} 

/**********************************************************************
 * SIR transmission
 **********************************************************************/

static void writeProtected(SerialDevice* sd, int c)
{
  if(c==SIR_BOF || c==SIR_EOF || c==SIR_CE) {
    sd->putChar(sd,SIR_CE);
    sd->putChar(sd,SIR_ESC_BIT^c);
  } else {
    sd->putChar(sd,c);
  }
}

static void sendFrame(FrameDevice* fd, const void* buf0, int len)
{
  u_char* buf=(u_char*)buf0;
  SIRFrameDevice* sfd=(SIRFrameDevice*)fd;
  SerialDevice* sd=sfd->serial;
  int i,ofcs=INITFCS;

  for(i=sfd->ebofs;i>0;i--) sd->putChar(sd,SIR_EXTRA_BOF);
  sd->putChar(sd,SIR_BOF);

  for(i=0;i<len;i++) {
    int c=buf[i];
    ofcs=updateFCS(ofcs,c);
    writeProtected(sd,c);
  }

  ofcs=~ofcs;
  writeProtected(sd,0xff&ofcs);
  writeProtected(sd,0xff&(ofcs>>8));

  sd->putChar(sd,SIR_EOF);
}

/**********************************************************************
 * SIR reception
 **********************************************************************/

static void timeoutListener(void* fd)
{
  SIRFrameDevice* sfd=(SIRFrameDevice*)fd;

  if((sfd->frame.debug&FRAME_DEBUG_INPUT) && sfd->frameState!=FRAME_OUTSIDE)
    log("input frame timeout\n");
  sfd->frameState=FRAME_OUTSIDE;
}

static void status(SerialDevice* sd, int event)
{
  SIRFrameDevice* sfd=(SIRFrameDevice*)sd->handle;

  if(event==SERIAL_INPUT_AVAILABLE && sfd->inBuf) {
    int c;

    while((c=sd->getChar(sd))!=SERIAL_INBUF_EMPTY) {
      switch(c) {
      case SIR_BOF:
	sfd->frameState=FRAME_INSIDE;
	sfd->inChars=0;
	sfd->inFCS=INITFCS;
	break;
      case SIR_EOF:
	if(sfd->frameState==FRAME_INSIDE && sfd->inChars>=4 && sfd->inFCS==GOODFCS) {
	  if(sfd->frame.frame)
	    sfd->frame.frame(&sfd->frame,sfd->inBuf,sfd->inChars-2);
	} else if((sfd->frame.debug&FRAME_DEBUG_INPUT) && sfd->frameState!=FRAME_OUTSIDE) {
	  log("malformed input frame\n");
	}
	sfd->frameState=FRAME_OUTSIDE;
	break;
      case SIR_CE:
	if(sfd->frameState==FRAME_INSIDE) sfd->frameState=FRAME_ESCAPE;
	break;
      default:
	if(sfd->frameState!=FRAME_OUTSIDE) {
	  if(sfd->frameState==FRAME_ESCAPE) {
	    sfd->frameState=FRAME_INSIDE;
	    c^=SIR_ESC_BIT;
	  }
	  if(sfd->inChars<sfd->maxSize+2) {
	    sfd->inBuf[sfd->inChars++]=c;
	    sfd->inFCS=updateFCS(sfd->inFCS,c);
	  } else {
	    sfd->frameState=FRAME_OUTSIDE;
	    if(sfd->frame.debug&FRAME_DEBUG_INPUT) log("input frame overrun\n");
	  }
	}
	break;
      }
      if(sfd->frameState!=FRAME_OUTSIDE) evtSetTimer(10,timeoutListener,sfd);
    }
  }
}

/**********************************************************************
 * Miscellaneous
 **********************************************************************/

static int getSpeedMask(FrameDevice* fd)
{
  SIRFrameDevice* sfd=(SIRFrameDevice*)fd;

  return sfd->serial->getSpeedMask(sfd->serial);
}

static int getMinTurnaroundMask(FrameDevice* fd)
{
  SIRFrameDevice* sfd=(SIRFrameDevice*)fd;

  return sfd->serial->getMinTurnaroundMask(sfd->serial);
}

static int setParams(FrameDevice* fd, int baud, int ebofs, int maxSize)
{
  SIRFrameDevice* sfd=(SIRFrameDevice*)fd;
  int delay=sfd->serial->setSpeed(sfd->serial,baud);

  sfd->ebofs=ebofs;
  if(sfd->maxSize!=maxSize) {
    sfd->maxSize=maxSize;
    if(sfd->inBuf) freeMem(sfd->inBuf);
    sfd->inBuf=sfd->maxSize ? allocMem(id_sir_inbuffer,maxSize+2) : 0;
  }    
  sfd->frameState=FRAME_OUTSIDE;

  return delay;
}

static void resetParams(FrameDevice* fd)
{
  SIRFrameDevice* sfd=(SIRFrameDevice*)fd;

  sfd->serial->setSpeed(sfd->serial,DEFAULT_SPEED);
  sfd->ebofs=DEFAULT_EBOFS;
  if(sfd->maxSize!=DEFAULT_SIZE) {
    sfd->maxSize=DEFAULT_SIZE;
    if(sfd->inBuf) freeMem(sfd->inBuf);
    sfd->inBuf=sfd->maxSize ? allocMem(id_sir_inbuffer,DEFAULT_SIZE+2) : 0;
  }    
  sfd->frameState=FRAME_OUTSIDE;
}

static void sirClose(FrameDevice* fd)
{
  SIRFrameDevice* sfd=(SIRFrameDevice*)fd;

  evtCancelTimer(timeoutListener,sfd);
  sfd->serial->handle=0;
  sfd->serial->status=0;
  if(sfd->inBuf) freeMem(sfd->inBuf);
  freeMem(sfd);
}


/**********************************************************************
 * External interface
 **********************************************************************/

FrameDevice* createSIRFrameDevice(SerialDevice* serial)
{
  SIRFrameDevice* sfd=allocMem(id_sir_framedevice,sizeof(SIRFrameDevice));

  sfd->frame.close=sirClose;
  sfd->frame.setParams=setParams;
  sfd->frame.resetParams=resetParams;
  sfd->frame.sendFrame=sendFrame;
  sfd->frame.getSpeedMask=getSpeedMask;
  sfd->frame.getMinTurnaroundMask=getMinTurnaroundMask;
  sfd->frame.debug=0;
  sfd->frame.handle=0;
  sfd->frame.frame=0;

  sfd->serial=serial;
  sfd->inBuf=0;
  sfd->maxSize=0;
  sfd->frameState=FRAME_OUTSIDE;
  sfd->ebofs=10;

  serial->handle=sfd;
  serial->status=status;

  return &sfd->frame;
}

#define ADDBYTE(x) \
    if (ix < maxlen) \
      dst[ix++] = (x); \
    else \
      goto bad

#define ADDPROTECTED(c) \
  if((c) == SIR_BOF || (c) == SIR_EOF || (c) == SIR_CE) { \
    ADDBYTE(SIR_CE); \
    ADDBYTE(SIR_ESC_BIT^(c)); \
  } else { \
    ADDBYTE(c); \
  }

int
createSIRFrame(int ebofs, const u_char *src, int len, u_char *dst, int maxlen)
{
  int i, ofcs = INITFCS;
  int ix = 0;

  for(i = ebofs; i > 0; i--)
    ADDBYTE(SIR_EXTRA_BOF);

  ADDBYTE(SIR_BOF);

  for(i = 0; i < len; i++) {
    int c = src[i];
    ofcs = updateFCS(ofcs, c);
    ADDPROTECTED(c);
  }

  ofcs=~ofcs;
  ADDPROTECTED(0xff&ofcs);
  ADDPROTECTED(0xff&(ofcs>>8));

  ADDBYTE(SIR_EOF);
  return ix;
 bad:
  log("buffer overrun in createSIRFrame: in size=%d, max size=%d\n", 
      len, maxlen);
  exit(-1);
}

==> birda-1.3.tar.gz: birda-1.3/src/sir.c <==
/*
 * Copyright (c) 2001 Tommy Bohlin <tommy@gatespace.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/* sir.c
 */

#include <irda.h>

#include <stdlib.h>

/**********************************************************************
 * Constants
 **********************************************************************/

static const char id_sir_framedevice[]="sir frame device";
static const char id_sir_inbuffer[]="sir inbuffer";

/**********************************************************************
 * Data structures
 **********************************************************************/

#define SIR_EXTRA_BOF            0xff
#define SIR_BOF                  0xc0
#define SIR_CE                   0x7d
#define SIR_EOF                  0xc1

#define SIR_ESC_BIT              0x20

typedef struct SIRFrameDevice {
  FrameDevice frame;
  SerialDevice* serial;

  u_char* inBuf;
  int maxSize;
  int frameState;
#define FRAME_OUTSIDE    0
#define FRAME_INSIDE     1
#define FRAME_ESCAPE     2
  int inChars;
  int inFCS;

  int ebofs;
} SIRFrameDevice;

/**********************************************************************
 * CRC
 **********************************************************************/

static const int fcstab[] = {
  0x0000, 0x1189, 0x2312, 0x329b, 0x4624, 0x57ad, 0x6536, 0x74bf,
  0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5, 0xe97e, 0xf8f7,
  0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7, 0x643e,
  0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
  0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd,
  0xad4a, 0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5,
  0x3183, 0x200a, 0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c,
  0xbdcb, 0xac42, 0x9ed9, 0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974,
  0x4204, 0x538d, 0x6116, 0x709f, 0x0420, 0x15a9, 0x2732, 0x36bb,
  0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868, 0x99e1, 0xab7a, 0xbaf3,
  0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528, 0x37b3, 0x263a,
  0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb, 0xaa72,
  0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
  0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1,
  0x7387, 0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738,
  0xffcf, 0xee46, 0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70,
  0x8408, 0x9581, 0xa71a, 0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7,
  0x0840, 0x19c9, 0x2b52, 0x3adb, 0x4e64, 0x5fed, 0x6d76, 0x7cff,
  0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad, 0xc324, 0xf1bf, 0xe036,
  0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c, 0x7df7, 0x6c7e,
  0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c, 0xd1b5,
  0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
  0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134,
  0x39c3, 0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c,
  0xc60c, 0xd785, 0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3,
  0x4a44, 0x5bcd, 0x6956, 0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb,
  0xd68d, 0xc704, 0xf59f, 0xe416, 0x90a9, 0x8120, 0xb3bb, 0xa232,
  0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1, 0x0d68, 0x3ff3, 0x2e7a,
  0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3, 0x8238, 0x93b1,
  0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70, 0x1ff9,
  0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
  0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78
};

static const int INITFCS=0xffff;
static const int GOODFCS=0xf0b8;

static int updateFCS(int fcs, int c) {
  return (fcs>>8)^fcstab[0xff&(fcs^c)];
} 

/**********************************************************************
 * SIR transmission
 **********************************************************************/

static void writeProtected(SerialDevice* sd, int c)
{
  if(c==SIR_BOF || c==SIR_EOF || c==SIR_CE) {
    sd->putChar(sd,SIR_CE);
    sd->putChar(sd,SIR_ESC_BIT^c);
  } else {
    sd->putChar(sd,c);
  }
}

static void sendFrame(FrameDevice* fd, const void* buf0, int len)
{
  u_char* buf=(u_char*)buf0;
  SIRFrameDevice* sfd=(SIRFrameDevice*)fd;
  SerialDevice* sd=sfd->serial;
  int i,ofcs=INITFCS;

  for(i=sfd->ebofs;i>0;i--) sd->putChar(sd,SIR_EXTRA_BOF);
  sd->putChar(sd,SIR_BOF);

  for(i=0;i<len;i++) {
    int c=buf[i];
    ofcs=updateFCS(ofcs,c);
    writeProtected(sd,c);
  }

  ofcs=~ofcs;
  writeProtected(sd,0xff&ofcs);
  writeProtected(sd,0xff&(ofcs>>8));

  sd->putChar(sd,SIR_EOF);
}

/**********************************************************************
 * SIR reception
 **********************************************************************/

static void timeoutListener(void* fd)
{
  SIRFrameDevice* sfd=(SIRFrameDevice*)fd;

  if((sfd->frame.debug&FRAME_DEBUG_INPUT) && sfd->frameState!=FRAME_OUTSIDE)
    birda_log("input frame timeout\n");
  sfd->frameState=FRAME_OUTSIDE;
}

static void status(SerialDevice* sd, int event)
{
  SIRFrameDevice* sfd=(SIRFrameDevice*)sd->handle;

  if(event==SERIAL_INPUT_AVAILABLE && sfd->inBuf) {
    int c;

    while((c=sd->getChar(sd))!=SERIAL_INBUF_EMPTY) {
      switch(c) {
      case SIR_BOF:
	sfd->frameState=FRAME_INSIDE;
	sfd->inChars=0;
	sfd->inFCS=INITFCS;
	break;
      case SIR_EOF:
	if(sfd->frameState==FRAME_INSIDE && sfd->inChars>=4 && sfd->inFCS==GOODFCS) {
	  if(sfd->frame.frame)
	    sfd->frame.frame(&sfd->frame,sfd->inBuf,sfd->inChars-2);
	} else if((sfd->frame.debug&FRAME_DEBUG_INPUT) && sfd->frameState!=FRAME_OUTSIDE) {
	  birda_log("malformed input frame\n");
	}
	sfd->frameState=FRAME_OUTSIDE;
	break;
      case SIR_CE:
	if(sfd->frameState==FRAME_INSIDE) sfd->frameState=FRAME_ESCAPE;
	break;
      default:
	if(sfd->frameState!=FRAME_OUTSIDE) {
	  if(sfd->frameState==FRAME_ESCAPE) {
	    sfd->frameState=FRAME_INSIDE;
	    c^=SIR_ESC_BIT;
	  }
	  if(sfd->inChars<sfd->maxSize+2) {
	    sfd->inBuf[sfd->inChars++]=c;
	    sfd->inFCS=updateFCS(sfd->inFCS,c);
	  } else {
	    sfd->frameState=FRAME_OUTSIDE;
	    if(sfd->frame.debug&FRAME_DEBUG_INPUT) birda_log("input frame overrun\n");
	  }
	}
	break;
      }
      if(sfd->frameState!=FRAME_OUTSIDE) evtSetTimer(10,timeoutListener,sfd);
    }
  }
}

/**********************************************************************
 * Miscellaneous
 **********************************************************************/

static int getSpeedMask(FrameDevice* fd)
{
  SIRFrameDevice* sfd=(SIRFrameDevice*)fd;

  return sfd->serial->getSpeedMask(sfd->serial);
}

static int getMinTurnaroundMask(FrameDevice* fd)
{
  SIRFrameDevice* sfd=(SIRFrameDevice*)fd;

  return sfd->serial->getMinTurnaroundMask(sfd->serial);
}

static int setParams(FrameDevice* fd, int baud, int ebofs, int maxSize)
{
  SIRFrameDevice* sfd=(SIRFrameDevice*)fd;
  int delay=sfd->serial->setSpeed(sfd->serial,baud);

  sfd->ebofs=ebofs;
  if(sfd->maxSize!=maxSize) {
    sfd->maxSize=maxSize;
    if(sfd->inBuf) freeMem(sfd->inBuf);
    sfd->inBuf=sfd->maxSize ? allocMem(id_sir_inbuffer,maxSize+2) : 0;
  }    
  sfd->frameState=FRAME_OUTSIDE;

  return delay;
}

static void resetParams(FrameDevice* fd)
{
  SIRFrameDevice* sfd=(SIRFrameDevice*)fd;

  sfd->serial->setSpeed(sfd->serial,DEFAULT_SPEED);
  sfd->ebofs=DEFAULT_EBOFS;
  if(sfd->maxSize!=DEFAULT_SIZE) {
    sfd->maxSize=DEFAULT_SIZE;
    if(sfd->inBuf) freeMem(sfd->inBuf);
    sfd->inBuf=sfd->maxSize ? allocMem(id_sir_inbuffer,DEFAULT_SIZE+2) : 0;
  }    
  sfd->frameState=FRAME_OUTSIDE;
}

static void sirClose(FrameDevice* fd)
{
  SIRFrameDevice* sfd=(SIRFrameDevice*)fd;

  evtCancelTimer(timeoutListener,sfd);
  sfd->serial->handle=0;
  sfd->serial->status=0;
  if(sfd->inBuf) freeMem(sfd->inBuf);
  freeMem(sfd);
}


/**********************************************************************
 * External interface
 **********************************************************************/

FrameDevice* createSIRFrameDevice(SerialDevice* serial)
{
  SIRFrameDevice* sfd=allocMem(id_sir_framedevice,sizeof(SIRFrameDevice));

  sfd->frame.close=sirClose;
  sfd->frame.setParams=setParams;
  sfd->frame.resetParams=resetParams;
  sfd->frame.sendFrame=sendFrame;
  sfd->frame.getSpeedMask=getSpeedMask;
  sfd->frame.getMinTurnaroundMask=getMinTurnaroundMask;
  sfd->frame.debug=0;
  sfd->frame.handle=0;
  sfd->frame.frame=0;

  sfd->serial=serial;
  sfd->inBuf=0;
  sfd->maxSize=0;
  sfd->frameState=FRAME_OUTSIDE;
  sfd->ebofs=10;

  serial->handle=sfd;
  serial->status=status;

  return &sfd->frame;
}

#define ADDBYTE(x) \
    if (ix < maxlen) \
      dst[ix++] = (x); \
    else \
      goto bad

#define ADDPROTECTED(c) \
  if((c) == SIR_BOF || (c) == SIR_EOF || (c) == SIR_CE) { \
    ADDBYTE(SIR_CE); \
    ADDBYTE(SIR_ESC_BIT^(c)); \
  } else { \
    ADDBYTE(c); \
  }

int
createSIRFrame(int ebofs, const u_char *src, int len, u_char *dst, int maxlen)
{
  int i, ofcs = INITFCS;
  int ix = 0;

  for(i = ebofs; i > 0; i--)
    ADDBYTE(SIR_EXTRA_BOF);

  ADDBYTE(SIR_BOF);

  for(i = 0; i < len; i++) {
    int c = src[i];
    ofcs = updateFCS(ofcs, c);
    ADDPROTECTED(c);
  }

  ofcs=~ofcs;
  ADDPROTECTED(0xff&ofcs);
  ADDPROTECTED(0xff&(ofcs>>8));

  ADDBYTE(SIR_EOF);
  return ix;
 bad:
  birda_log("buffer overrun in createSIRFrame: in size=%d, max size=%d\n", 
      len, maxlen);
  exit(-1);
}
